@@ -1,5 +1,5 @@
 #  This file is part of GNU Dico.
-#  Copyright (C) 2008-2010, 2012 Wojciech Polak
+#  Copyright (C) 2008-2010, 2012, 2013, 2014 Wojciech Polak
 #
 #  GNU Dico is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -15,13 +15,15 @@
 #  along with GNU Dico.  If not, see <http://www.gnu.org/licenses/>.
 
 from django.conf import settings
-from django.conf.urls.defaults import *
-import views
+from django.conf.urls import url, include
+# Uncomment this when running via runserver
+#from django.conf.urls.static import static
+from . import views as app_views
+
+urlpatterns = [
+    url(r'^$', app_views.index, name='index'),
+    url(r'^opensearch\.xml$', app_views.opensearch, name='opensearch'),
+]
+# Add the following when running via runserver:
+#  + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
 
-urlpatterns = patterns (
-    '',
-    (r'^/?$', views.index, {}, 'index'),
-    (r'^opensearch\.xml$', views.opensearch, {}, 'opensearch'),
-    (r'^static/(?P<path>.*)$', 'django.views.static.serve',
-     {'document_root': settings.MEDIA_ROOT}),
-)
