@@ -1,5 +1,5 @@
 /* grecs - Gray's Extensible Configuration System
-   Copyright (C) 2007-2012 Sergey Poznyakoff
+   Copyright (C) 2007-2021 Sergey Poznyakoff
 
    Grecs is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -27,10 +27,13 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <limits.h>
+#include <string.h>
 #include "grecs.h"
 
+struct grecs_sockaddr_hints *grecs_sockaddr_hints;
+
 void
-grecs_value_free(struct grecs_value *val)
+grecs_value_free_content(struct grecs_value *val)
 {
 	int i;
 
@@ -50,9 +53,28 @@ grecs_value_free(struct grecs_value *val)
 			grecs_value_free(val->v.arg.v[i]);
 		free(val->v.arg.v);
 	}
+}
+
+void
+grecs_value_free(struct grecs_value *val)
+{
+	grecs_value_free_content(val);
 	grecs_free(val);
 }
 
+const char *
+grecs_value_type_string(int t)
+{
+	static const char *types[] = {
+		[GRECS_TYPE_STRING] = N_("string"),
+		[GRECS_TYPE_LIST]   = N_("list"),
+		[GRECS_TYPE_ARRAY]  = N_("one or more arguments")
+	};
+	if (t >= 0 && t < sizeof(types) / sizeof(types[0]))
+		return types[t];
+	return N_("unrecognized type; please report");
+}
+
 struct grecs_node *
 grecs_node_create(enum grecs_node_type type, grecs_locus_t *loc)
 {
@@ -119,9 +141,9 @@ grecs_node_unlink(struct grecs_node *node)
 		return 1;
 	if (node->next)
 		node->next->prev = node->prev;
+	node->up = node->prev = node->next = NULL;
 	return 0;
 }
-
 
 static void
 listel_dispose(void *el)
@@ -234,7 +256,6 @@ grecs_tree_free(struct grecs_node *node)
 {
 	if (!node)
 		return 0;
-        /* FIXME: Check if that's a node_type_root */
 	if (node->type != grecs_node_root) {
 		errno = EINVAL;
 		return 1;
@@ -246,11 +267,20 @@ grecs_tree_free(struct grecs_node *node)
 
 
 static int
-fake_callback(enum grecs_callback_command cmd,
-	       grecs_locus_t *locus,
-	       void *varptr,
-	       grecs_value_t *value,
-	       void *cb_data)
+fake_callback(
+#if GRECS_TREE_API
+	enum grecs_callback_command cmd,
+	grecs_node_t *node,
+	void *varptr,
+	void *cb_data
+#else	
+	enum grecs_callback_command cmd,
+	grecs_locus_t *locus,
+	void *varptr,
+	grecs_value_t *value,
+	void *cb_data
+#endif	
+	)
 {
 	return 0;
 }
@@ -269,14 +299,32 @@ static struct grecs_keyword fake = {
 };
 
 static struct grecs_keyword *
-find_keyword(struct grecs_keyword *cursect, const char *ident)
+find_keyword(struct grecs_keyword *cursect, grecs_node_t *node)
 {
-	struct grecs_keyword *kwp;
-  
 	if (cursect && cursect->kwd && cursect != &fake) {
-		for (kwp = cursect->kwd; kwp->ident; kwp++)
-			if (strcmp(kwp->ident, ident) == 0)
-				return kwp;
+		struct grecs_keyword *found = NULL, *kwp;
+		char const *msg;
+
+		for (kwp = cursect->kwd; kwp->ident; kwp++) {
+			if (strcmp(kwp->ident, node->ident) == 0) {
+				found = kwp;
+				if (kwp->callback
+				    || (node->down
+				         ? kwp->type == grecs_type_section
+					: kwp->type != grecs_type_section))
+					return kwp;
+			}
+		}
+		if (found) {
+			if (found->type == grecs_type_section) {
+				msg = N_("section keyword used as a scalar");
+			} else {
+				msg = N_("scalar keyword used as a section");
+			}
+		} else {
+			msg = N_("unknown keyword");
+		}
+		grecs_error(&node->idloc, 0, "%s", gettext(msg));
 	} else {
 		return &fake;
 	}
@@ -331,6 +379,7 @@ string_to_host(struct in_addr *in, const char *string,
 	return 0;
 }
 
+#if !GRECS_SOCKADDR_LIST
 static int
 string_to_sockaddr(struct grecs_sockaddr *sp, const char *string,
 		   grecs_locus_t const *locus)
@@ -409,7 +458,7 @@ string_to_sockaddr(struct grecs_sockaddr *sp, const char *string,
 	}
 	return 0;
 }
-
+#endif
 
 /* The TYPE_* defines come from gnulib's intprops.h */
 
@@ -528,14 +577,17 @@ string_to_sockaddr(struct grecs_sockaddr *sp, const char *string,
 
 int
 grecs_string_convert(void *target, enum grecs_data_type type,
-		      const char *string, grecs_locus_t const *locus)
+		     const char *string, grecs_locus_t const *locus)
 {
   switch (type) {
   case grecs_type_void:
 	  abort();
-	    
+
+  case grecs_type_null:
+	  break;
+	  
   case grecs_type_string:
-	  *(const char**)target = grecs_strdup(string);
+	  *(char**)target = grecs_strdup(string);
 	  break;
 	    
   case grecs_type_short:
@@ -596,14 +648,17 @@ grecs_string_convert(void *target, enum grecs_data_type type,
 	  break;    
 	    
   case grecs_type_sockaddr:
+#if GRECS_SOCKADDR_LIST
+	  return grecs_str_to_sockaddr((struct grecs_sockaddr **)target,
+				       string, grecs_sockaddr_hints,
+				       locus);
+#else
 	  return string_to_sockaddr((struct grecs_sockaddr *)target, string,
 				    locus);
-      
-      /* FIXME: */
+#endif
+
   case grecs_type_cidr:
-	  grecs_error(locus, 0,
-		       _("INTERNAL ERROR at %s:%d"), __FILE__, __LINE__);
-	  abort();
+	  return grecs_str_to_cidr((struct grecs_cidr *)target, string, locus);
 	    
   case grecs_type_section:
 	  grecs_error(locus, 0,
@@ -644,9 +699,19 @@ DECL_NUMCMP(time_t)
 __DECL_NUMCMP(in_addr, struct in_addr)
 __DECL_NUMCMP(grecs_sockaddr, struct grecs_sockaddr)
 
+static int
+cidr_cmp(const void *elt1, const void *elt2)
+{
+	struct grecs_cidr const *cp1 = elt1, *cp2 = elt2;
+	return !(cp1->family == cp2->family
+		 && cp1->len == cp2->len
+		 && memcmp(cp1->address, cp2->address, cp1->len) == 0
+		 && memcmp(cp1->netmask, cp2->netmask, cp1->len) == 0);
+}
+
 struct grecs_prop grecs_prop_tab[] = {
-	{ 0, NULL },                                  /* grecs_type_void */
-	{ sizeof(char*), string_cmp },                /* grecs_type_string */
+	{ 0, NULL },                                 /* grecs_type_void */
+	{ sizeof(char*), string_cmp },               /* grecs_type_string */
 	{ sizeof(short), NUMCMP(short) },            /* grecs_type_short */
 	{ sizeof(unsigned short), NUMCMP(short) },   /* grecs_type_ushort */
 	{ sizeof(int), NUMCMP(int) },                /* grecs_type_int */
@@ -661,11 +726,12 @@ struct grecs_prop grecs_prop_tab[] = {
 	{ sizeof(time_t), NUMCMP(time_t) },          /* grecs_type_time */
 	{ sizeof(int), NUMCMP(int) },                /* grecs_type_bool */
 	{ sizeof(struct in_addr), NUMCMP(in_addr) }, /* grecs_type_ipv4 */
-	{ 0, NULL },                           /* FIXME: grecs_type_cidr */
+	{ sizeof(struct grecs_cidr), cidr_cmp },     /* grecs_type_cidr */
 	{ sizeof(struct in_addr), NUMCMP(in_addr) }, /* grecs_type_host */ 
 	{ sizeof(struct grecs_sockaddr), NUMCMP(grecs_sockaddr) },
                                                	      /* grecs_type_sockaddr */
-	{ 0, NULL }                                   /* grecs_type_section */
+	{ 0, NULL },                                  /* grecs_type_section */
+	{ 0, NULL }                                   /* grecs_type_null */
 };
 #define grecs_prop_count \
  (sizeof(grecs_prop_tab) / sizeof(grecs_prop_tab[0]))
@@ -681,13 +747,24 @@ grecs_process_ident(struct grecs_keyword *kwp, grecs_value_t *value,
 	
 	target = target_ptr(kwp, (char *) base);
 	
-	if (kwp->callback)
+	if (kwp->callback) {
+#if GRECS_TREE_API
+		struct grecs_node node = { 0 };
+		node.locus = *locus;
+		node.v.value = value;
+		node.ident = (char*) kwp->ident;
+		kwp->callback(grecs_callback_set_value,
+			      &node,
+			      target,
+			      &kwp->callback_data);
+#else
 		kwp->callback(grecs_callback_set_value,
 			      locus,
 			      target,
 			      value,
 			      &kwp->callback_data);
-	else if (kwp->type == grecs_type_void || target == NULL)
+#endif
+	} else if (kwp->type == grecs_type_void || target == NULL)
 		return;
 	else if (!value) {
 		grecs_error(locus, 0, "%s has no value", kwp->ident);
@@ -778,10 +855,15 @@ grecs_process_ident(struct grecs_keyword *kwp, grecs_value_t *value,
 			grecs_list_append(list, ptr);
 		}
 		*(struct grecs_list**)target = list;
-	} else
+	} else {
+		if (kwp->type == grecs_type_string
+		    && !(kwp->flags & GRECS_CONST))
+			free(*(char**)target);
 		grecs_string_convert(target, kwp->type,
 				     value->v.string,
 				     &value->locus);
+	}
+	kwp->flags &= ~GRECS_CONST;
 }
 
 
@@ -804,11 +886,18 @@ stmt_begin(struct nodeproc_closure *clos,
 		target = target_ptr(kwp, CURRENT_BASE(clos));
 		clos->cursect = kwp;
 		if (kwp->callback) {
+#if GRECS_TREE_API			
+			if (kwp->callback(grecs_callback_section_begin,
+					  node,
+					  target,
+					  &kwp->callback_data))
+#else
 			if (kwp->callback(grecs_callback_section_begin,
 					  &node->locus,
 					  target,
 					  node->v.value,
 					  &kwp->callback_data))
+#endif
 				clos->cursect = &fake;
 		} else
 			kwp->callback_data = target;
@@ -833,11 +922,18 @@ stmt_end(struct nodeproc_closure *clos, struct grecs_node *node)
 	if (!clos->cursect)
 		abort();
 	if (callback)
+#if GRECS_TREE_API			
+		callback(grecs_callback_section_end,
+			 node,
+			 kwp ? target_ptr(kwp, CURRENT_BASE(clos)) : NULL,
+			 dataptr);
+#else		
 		callback(grecs_callback_section_end,
 			 &node->locus,
 			 kwp ? target_ptr(kwp, CURRENT_BASE(clos)) : NULL,
 			 NULL,
 			 dataptr);
+#endif
 	if (kwp)
 		kwp->callback_data = NULL;
 }
@@ -850,21 +946,17 @@ nodeproc(enum grecs_tree_recurse_op op, struct grecs_node *node, void *data)
 	
 	switch (op) {
 	case grecs_tree_recurse_set:
-		kwp = find_keyword(clos->cursect, node->ident);
-		if (!kwp) {
-			grecs_error(&node->idloc, 0, _("Unknown keyword"));
+		kwp = find_keyword(clos->cursect, node);
+		if (!kwp)
 			return grecs_tree_recurse_skip;
-		}
 		grecs_process_ident(kwp, node->v.value, CURRENT_BASE(clos),
 				    &node->idloc);
 		break;
 		
 	case grecs_tree_recurse_pre:
-		kwp = find_keyword(clos->cursect, node->ident);
-		if (!kwp) {
-			grecs_error(&node->locus, 0, _("Unknown keyword"));
+		kwp = find_keyword(clos->cursect, node);
+		if (!kwp)
 			return grecs_tree_recurse_skip;
-		}
 		stmt_begin(clos, kwp, node);
 		break;
 		
@@ -1025,7 +1117,10 @@ node_merge_block(struct grecs_node *to_node, struct grecs_node *from_node,
 		if (!sp->next)
 			break;
 	}
+
 	sp->next = to_node->down;
+	to_node->down->prev = sp;
+	
 	to_node->down = from_node->down;
 	from_node->down = NULL;
 }
@@ -1041,7 +1136,7 @@ node_reduce(struct grecs_node *node, struct grecs_keyword *kwp, int flags)
 	if (p) {
 		switch (node->type) {
 		case grecs_node_root:
-			return;
+			return 0;
 		case grecs_node_stmt:
 			node_merge_stmt(p, node, kwp, flags);
 			break;
@@ -1068,13 +1163,9 @@ reduceproc(enum grecs_tree_recurse_op op, struct grecs_node *node, void *data)
 	} else {
 		struct grecs_keyword *kwp = NULL;
 		if (clos->cursect) {
-			kwp = find_keyword(clos->cursect, node->ident);
-			if (!kwp) {
-				grecs_error(&node->locus, 0,
-					    _("%s: unknown keyword"),
-					    node->ident);
+			kwp = find_keyword(clos->cursect, node);
+			if (!kwp)
 				return grecs_tree_recurse_skip;
-			}
 			if (kwp->flags & GRECS_INAC)
 				return grecs_tree_recurse_skip;
 			if (!(kwp->flags & GRECS_MULT) &&
@@ -1117,12 +1208,9 @@ grecs_tree_reduce(struct grecs_node *node, struct grecs_keyword *kwd,
 struct grecs_node *
 grecs_tree_first_node(struct grecs_node *tree)
 {
-	if (tree->type != grecs_node_root) {
-		errno = EINVAL;
-		return NULL;
-	}
-	errno = 0;
-	return tree->down;
+	if (tree->type == grecs_node_root)
+		return tree->down;
+	return tree;
 }
 
 struct grecs_node *
