@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2018 Sergey Poznyakoff
+   Copyright (C) 1998-2021 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,6 +22,12 @@
 #endif
 #include <dico.h>
 
+/* Starting from gnulib commit 52ab50a7ebc469b542bca0063430db71aefb67af,
+   xalloc.h declares xalloc_die only if GNULIB_XALLOC_DIE is defined.
+   Without this define each inclusion of xalloc.h will spit out one or
+   more warnings about this function not being declared. */
+#define GNULIB_XALLOC_DIE 1
+#include "xalloc.h"
 
 struct sockaddr;
 void sockaddr_to_str(const struct sockaddr *sa, int salen,
