@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico
-   Copyright (C) 2007-2008, 2010, 2012, 2016-2018 Sergey Poznyakoff
+   Copyright (C) 2007-2021 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,9 +14,7 @@
    You should have received a copy of the GNU General Public License
    along with GNU Dico.  If not, see <http://www.gnu.org/licenses/>. */
 
-#ifdef HAVE_CONFIG_H
-# include <config.h>
-#endif
+#include <config.h>
 #include <dico.h>
 #include <limits.h>
 #include <errno.h>
@@ -1540,9 +1538,19 @@ utf8_strlen(const char *s)
 static int
 utf8_iter0(struct utf8_iterator *itr)
 {
-    size_t n = utf8_char_width(itr->curptr);
-    if (n == 0)
+    size_t n;
+
+    if (itr->length == 0) {
+	itr->end = 1;
 	return 1;
+    }
+    n = utf8_char_width(itr->curptr);
+    if (n == 0 || n > itr->length) {
+	itr->end = 1;
+	itr->err = 1;
+	return 1;
+    }
+    
     itr->curwidth = n;
     return 0;
 }
@@ -1550,23 +1558,40 @@ utf8_iter0(struct utf8_iterator *itr)
 int
 utf8_iter_end_p(struct utf8_iterator *itr)
 {
-    return *itr->curptr == 0;
+    return itr->end;
+}
+
+int
+utf8_iter_err_p(struct utf8_iterator *itr)
+{
+    return itr->end;
+}
+
+int
+utf8_iter_init(struct utf8_iterator *itr, char *ptr, size_t size)
+{
+    itr->string = ptr;
+    itr->curptr = ptr;
+    itr->length = size;
+    itr->curwidth = 0;
+    itr->err = 0;
+    itr->end = 0;
+    return utf8_iter0(itr);
 }
 
 int
 utf8_iter_first(struct utf8_iterator *itr, char *ptr)
 {
-    itr->string = ptr;
-    itr->curptr = ptr;
-    return utf8_iter0(itr);
+    return utf8_iter_init(itr, ptr, strlen(ptr));
 }
 
 int
 utf8_iter_next(struct utf8_iterator *itr)
 {
-    if (*itr->curptr == 0)
+    if (utf8_iter_end_p(itr))
 	return -1;
     itr->curptr += itr->curwidth;
+    itr->length -= itr->curwidth;
     return utf8_iter0(itr);
 }
 
@@ -2074,6 +2099,8 @@ utf8_wc_strstr(const unsigned *text, const unsigned *pattern)
     size_t pattern_len = utf8_wc_strlen(pattern);
     size_t *nextab;
 
+    errno = 0;
+
     /* Handle corner cases */
     if (pattern_len > text_len)
 	return NULL;
