@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008-2018 Sergey Poznyakoff
+   Copyright (C) 2008-2021 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -53,9 +53,7 @@
 */   
 
 
-#ifdef HAVE_CONFIG_H
-# include <config.h>
-#endif
+#include <config.h>
 #include <dico.h>
 #include <string.h>
 #include <stdlib.h>
@@ -267,7 +265,7 @@ revert_word(char *dst, const char *src, size_t len)
     char *p = dst + len;
 
     *p = 0;
-    for (utf8_iter_first(&itr, (char *)src);
+    for (utf8_iter_init(&itr, (char *)src, len);
 	 !utf8_iter_end_p(&itr);
 	 utf8_iter_next(&itr)) {
 	p -= itr.curwidth;
